import bisect
from typing import Optional

from flexi_nlp_tools.flexi_dict.search_engine.fuzzy_search_result import FuzzySearchResult

from .search_index import SearchIndex
from .config import DEFAULT_QUERY_TRANSFORMATION_PRICE
from ..flexi_dict.exceptions import InvalidScoresError
from .exceptions import EmptyQueryError


def fuzzy_search(
        query: str,
        search_index: SearchIndex,
        topn: Optional[int] = None,
        query_transformation_price: float = DEFAULT_QUERY_TRANSFORMATION_PRICE
):
    fuzzy_search_result = fuzzy_search_internal(
        query=query,
        search_index=search_index,
        topn=topn,
        query_transformation_price=query_transformation_price)

    return [item.value for item in fuzzy_search_result]


def fuzzy_search_internal(
        query: str,
        search_index: SearchIndex,
        topn: Optional[int] = None,
        query_transformation_price: float = DEFAULT_QUERY_TRANSFORMATION_PRICE
):
    if not 0 <= query_transformation_price <= 1:
        raise InvalidScoresError([query_transformation_price, ])

    if not query:
        raise EmptyQueryError()

    search_result = list()
    added = set()

    for rate in sorted(search_index):
        n = 0
        rate_search_result = search_index[rate].search_internal(query=query)

        for item in rate_search_result:

            if item.value in added:
                continue

            added.add(item.value)
            if rate:
                # print(item.path, item.total_corrections_price)
                total_correction_price = __update_total_corrections_price(
                    item=item, rate=rate, query=query, query_transformation_price=query_transformation_price)
                item.set_total_correction_price(total_correction_price)

            pos = bisect.bisect_left(
                search_result, item.total_corrections_price, lo=0, hi=len(search_result),
                key=lambda x: x.total_corrections_price)
            search_result.insert(pos, item)
            n += 1
            if topn and n > topn:
                break

    if search_result:
        return search_result[:topn]

    return search_result


def __update_total_corrections_price(
        item: FuzzySearchResult, query: str, rate: int, query_transformation_price: float
):
    total_corrections_price = item.total_corrections_price
    if rate:
        total_corrections_price += query_transformation_price
        total_corrections_price += 1e-5 * __get_path_key_dist(item.path, item.key, query)

    return total_corrections_price


def __get_path_key_dist(path: str, key: str, query: str) -> float:
    """
    Посчитать цену разницы между длинами запроса и найденного пути для ранжирования результата поиска.

    Чем больше разница в длине - тем ниже опускаем в результате поиска.
    Ключ - это запрос с исправлениями, по которому нашли значение
    Путь - это путь в дереве до листка дереве, в котором лежит найденное значение

    Например, ключем может быть "ман" а путем - "мандарин" или "манка"

    :param str path: путь в дереве, по которому был найдено значение
    :param str key: ключ, по которому был найдено значение

    :return float: нормализированная цена разницы между длинами

    """
    # абсолютная разница между длиной пути и длиной ключа
    # если длина пути меньше длины запроса, то удваиваем путь;
    # например, если искали "ябл", а нашли "я" и "яблок", то "я" нужно опустиь ниже по рангу
    len_path, len_key, len_query = len(path), len(key), len(query)
    dist = len_path - len_key if len_path >= len_query else (len_query - len_path) ** 3

    # абсолютная разница между длиной последнего токена
    start = key.rfind(' ') + 1 if ' ' in key else 0
    len_path_last_token = (path[start:] + ' ').find(' ')
    len_key_last_token = len(key) - start
    len_token_dist = abs(len_path_last_token - len_key_last_token)

    # комбинируем расстояния
    if len_path >= len_query and not len_token_dist:
        dist = 0

    return dist



from typing import Dict
from dataclasses import dataclass

from ..flexi_dict import FlexiDict
from ..flexi_dict.search_engine.correction import SymbolInsertion
from ..flexi_dict.search_engine import SearchEngine


@dataclass
class SearchIndex(Dict[int, 'FlexiDict']):

    def __getitem__(self, key: int) -> 'FlexiDict':
        if key not in self:
            search_engine = SearchEngine(symbol_insertion=SymbolInsertion(price=1e-5))
            self[key] = FlexiDict(search_engine=search_engine)
        return super().__getitem__(key)



import re

from .tokenizer import tokenize, detokenize


CONSONANTS_EN = 'BCDFGHKLMNPQRSTVWXZ'


REGEX_RULES_EN2UK = [
    (rf'([{CONSONANTS_EN}][{CONSONANTS_EN}]+)I([{CONSONANTS_EN}]+)', r'\1І\2'),
    (rf'^([{CONSONANTS_EN}]+)I$', r'\1АЙ'),
    (rf':::U([{CONSONANTS_EN}][{CONSONANTS_EN}]+)', r'А\1'),
    (rf':::([{CONSONANTS_EN}])U([{CONSONANTS_EN}][{CONSONANTS_EN}]+)', r'\1А\2'),
    (r'([CGHMPTF])U', r'\1U'),
    (r':::U', r'Ю'),
    (r'([AE])H:::', r'\1'),
    (rf':::([{CONSONANTS_EN}]+)E:::', r'\1І'),
    ('IE:::', 'АЙ'),
    ('([IE])CIAN', r'\1ШН'),
    ('([IE])CIOU', r'\1Ш'),
    ('([IE])C', r'\1Ч'),
]


STATIC_RULES_EN2UK = {
    r'TIONS:::': 'ШНС',
    r'TION:::': 'ШН',
    r'SIONS:::': 'ЖНС',
    r'SION:::': 'ЖН',
    r'CIANS:::': 'ШНС',
    r'CIAN:::': 'ШН',
    r'OUGH:::': 'У',
    r'IGHTS:::': 'АЙТС',
    r'IGHT:::': 'АЙТ',
    r'CIOUS:::': 'ШС',
    r'NCE:::': 'НС',
    'SES:::': 'СЕС',
    'XES:::': 'КСЕС',
    'IES:::': 'АЙС',
    'ATE:::': 'ЕТ',
    'ATES:::': 'ЕТС',
    'AY:::': 'АЙ',
    'OY:::': 'ОЙ',
    'Y:::': 'І',
    'YS:::': 'ЙС',
    'ES:::': 'С',
    'E:::': '',
    'S:::': 'С',
    'C:::': 'К',
    ':::WH': 'В',
    ':::GH': 'Г',
    ':::SCI': 'САЙ',
    'CE': 'СE',
    'СI': 'СI',
    'СY': 'СY',
    "UE": 'У',
    "UI": 'У',
    'UO': 'У',
    'SHCH': 'Щ',
    'SC': 'Ш',
    'KH': 'Х',
    'TS': 'Ц',
    'THR': 'ТР',
    'TH': 'С',
    'CC': 'КЦ',
    'CH': 'Ч',
    'SH': 'Ш',
    "ZH": "Ж",
    "ZJ": "Ж",
    "IU": "Ю",
    "IA": "Я",
    "IO": "ЙО",
    "IE": "Є",
    "IY": "ІЙ",
    "II": "ІЙ",
    "YA": "Я",
    "YI": "ИЙ",
    "YU": "Ю",
    "YO": "ЙО",
    "YE": "Є",
    "OO": 'У',
    'EE': 'І',
    'PH': 'Ф',
    'GH': 'Ж',
    'CK': 'К',
    'EA': 'ІА',
    "A": "А",
    "B": "Б",
    "V": "В",
    "G": "Г",
    "D": "Д",
    "E": "Е",
    "Z": "З",
    "Y": "У",
    "K": "К",
    "L": "Л",
    "M": "М",
    "N": "Н",
    "O": "О",
    "P": "П",
    "R": "Р",
    "S": "С",
    "T": "Т",
    "U": "У",
    "F": "Ф",
    "Q": "К’Ю",
    "W": "В",
    "I": "І",
    "H": "Х",
    "J": "Ж",
    "X": "КС",
    "C": "К",
    '’': 'Ь',
    '`': 'Ь'
}


STATIC_RULES_UK2RU = {
    'Є': 'Е',
    'Е': 'Е',
    'И': 'Ы',
    'І': 'И',
    'Ґ': 'Г',
    'Ї': 'ЙИ',
    '’': 'Ь',
    '`': 'Ь'
}

REGEX_RULES_EN2UK_PATTERNS = [
    (re.compile(pattern, flags=re.IGNORECASE), replacement)
    for pattern, replacement in REGEX_RULES_EN2UK
]

STATIC_RULES_EN2UK_PATTERNS = {
    re.compile(pattern, flags=re.IGNORECASE): replacement
    for pattern, replacement in STATIC_RULES_EN2UK.items()
}

STATIC_RULES_UK2RU_PATTERNS = {
    re.compile(pattern, flags=re.IGNORECASE): replacement
    for pattern, replacement in STATIC_RULES_UK2RU.items()
}


def en2uk_translit(s: str) -> str:
    tokens, sub = tokenize(s)
    tokens_translit = [_en2uk_translit(token) for token in tokens]
    s_translit = detokenize(tokens_translit, sub)
    return s_translit


def uk2ru_translit(s: str) -> str:
    tokens, sub = tokenize(s)
    tokens_translit = [_uk2ru_translit(token) for token in tokens]
    s_translit = detokenize(tokens_translit, sub)
    return s_translit


def en2ru_translit(s: str) -> str:
    s_translit = en2uk_translit(s)
    return uk2ru_translit(s_translit)


def __preserve_case(word, replacement):

    match word:

        case _ if not word.isalpha():
            result = replacement.lower()
        case _ if word.islower():
            result = replacement.lower()
        case _ if word.isupper():
            result = replacement.upper()
        case _ if word.istitle():
            result = replacement.capitalize()
        case _:
            result = replacement.lower()

    return rf'{result}'


def _en2uk_translit(token: str) -> str:
    s_translit = ':::' + token + ':::'

    for pattern, replacement in REGEX_RULES_EN2UK_PATTERNS:
        s_translit = pattern.sub(
            lambda match: __preserve_case(
                match.group(),
                pattern.sub(replacement, match.group())),
            s_translit)

    for pattern, replacement in STATIC_RULES_EN2UK_PATTERNS.items():
        s_translit = pattern.sub(lambda match: __preserve_case(match.group(), replacement), s_translit)

    return s_translit.replace(':::', '').strip()


def _uk2ru_translit(token: str) -> str:
    def repl_static(match, repl):
        return __preserve_case(match.group(0), repl)

    s_translit = token
    for pattern, replacement in STATIC_RULES_UK2RU_PATTERNS.items():
        s_translit = pattern.sub(lambda x: repl_static(x, replacement), s_translit)

    return s_translit



class EmptyQueryError(ValueError):
    """
    Exception raised when input query is empty.

    Args:
        message (str, optional): Additional error message. Defaults to "Input query is empty."
    """
    def __init__(self, message=None):
        if message is None:
            message = "Input query is empty."
        super().__init__(message)



from typing import List, Tuple, Callable, Optional, Union
from itertools import chain
import logging

from .utils import tokenize, detokenize
from .config import MIN_START_TOKEN_LENGTH
from .search_index import SearchIndex


logger = logging.getLogger(__name__)


def build_search_index(
        data: List[Tuple[Union[int, str], str]],
        min_start_token_length: int = MIN_START_TOKEN_LENGTH,
        callback: Optional[Callable] = None,
        should_stop: Optional[Callable] = None,
        callback_step: int = 10000
) -> SearchIndex:
    search_index = SearchIndex()
    update_search_index(
        search_index=search_index,
        data=data,
        min_start_token_length=min_start_token_length,
        callback=callback,
        should_stop=should_stop,
        callback_step=callback_step)
    return search_index


def update_search_index(
        search_index: SearchIndex,
        data: List[Tuple[int, str]],
        min_start_token_length: int = MIN_START_TOKEN_LENGTH,
        callback: Optional[Callable] = None,
        should_stop: Optional[Callable] = None,
        callback_step: int = 10000
):

    for i, (idx, key) in enumerate(data):
        if i % callback_step == 0:

            if callback:
                callback(i / len(data))

            if should_stop and should_stop():
                break

        search_index[0][key.lower()] = idx

        tokens, seps = tokenize(key)
        if len(seps) == len(tokens):
            seps[-1] += ' '
        else:
            seps.append(' ')
        for i_token in range(1, len(tokens)):

            if len(tokens[i_token]) < min_start_token_length:
                continue

            alt_tokens = list(chain(tokens[i_token:], tokens[:i_token]))
            alt_seps = list(chain(seps[i_token:], seps[:i_token]))
            alt_key = detokenize(alt_tokens, alt_seps).strip()
            search_index[1][alt_key.lower()] = idx




from .flexi_dict import FlexiDict
from .search_engine import SearchEngine
from .search_engine.correction import Correction
from .utils import calculate_symbols_distances, calculate_symbols_weights

__all__ = ('FlexiDict', 'SearchEngine', 'Correction', 'calculate_symbols_distances', 'calculate_symbols_weights')
